@@ -3,7 +3,7 @@ package Net::Async::Github;
 use strict;
 use warnings;
 
-our $VERSION = '0.003';
+our $VERSION = '0.004';
 
 use parent qw(IO::Async::Notifier);
 
@@ -21,6 +21,10 @@ Net::Async::Github - support for L<https://github.com>'s REST API with L<IO::Asy
    token => '...',
   )
  );
+ # Give 'secret_team' pull access to all private repos
+ $gh->repos(visibility => 'private')
+    ->grant_team(secret_team => 'pull')
+    ->await;
 
 =head1 DESCRIPTION
 
@@ -32,6 +36,8 @@ no indirect;
 
 use Future;
 use Dir::Self;
+use Path::Tiny;
+use File::ShareDir;
 use URI;
 use URI::QueryParam;
 use URI::Template;
@@ -49,7 +55,9 @@ use Log::Any qw($log);
 
 use Net::Async::Github::Branch;
 use Net::Async::Github::User;
+use Net::Async::Github::Team;
 use Net::Async::Github::Plan;
+use Net::Async::Github::PullRequest;
 use Net::Async::Github::Repository;
 use Net::Async::Github::RateLimit;
 
@@ -189,13 +197,24 @@ sub pr {
     $uri->path(
         join '/', 'repos', $args{owner}, $args{repo}, 'pulls', $args{id}
     );
-    $self->request(
-        GET => $uri,
-        user => $self->api_key,
-        pass => '',
-        headers => {
-            'Accept' => 'application/vnd.github.v3.full+json',
-        },
+    $self->http_get(
+        uri => $uri,
+    )->transform(
+        done => sub {
+            Net::Async::Github::PullRequest->new(
+                %{$_[0]},
+                github => $self,
+            )
+        }
+    )
+}
+
+sub teams {
+    my ($self, %args) = @_;
+	$self->validate_args(%args);
+    $self->api_get_list(
+        uri   => $self->endpoint('team', org => $args{organisation}),
+        class => 'Net::Async::Github::Team',
     )
 }
 
@@ -209,6 +228,76 @@ sub Net::Async::Github::Repository::branches {
     )
 }
 
+sub Net::Async::Github::Repository::grant_team {
+    my ($self, %args) = @_;
+    my $gh = $self->github;
+	$gh->validate_args(%args);
+    $self->github->http_put(
+        uri => $self->github->endpoint(
+            'team_repo',
+            team  => $args{team},
+            owner => $self->owner->{login},
+            repo  => $self->name,
+        ),
+        data => {
+            permission => $args{permission},
+        },
+    )
+}
+
+sub Net::Async::Github::PullRequest::owner { shift->{base}{repo}{owner}{login} }
+sub Net::Async::Github::PullRequest::repo { shift->{base}{repo}{name} }
+sub Net::Async::Github::PullRequest::branch_name { shift->{head}{ref} }
+
+sub Net::Async::Github::PullRequest::merge {
+    my ($self, %args) = @_;
+    my $gh = $self->github;
+	$gh->validate_args(%args);
+    die 'invalid owner' if ref $self->owner;
+    die 'invalid repo' if ref $self->repo;
+    die 'invalid id' if ref $self->id;
+    my $uri = $gh->endpoint(
+        'pull_request_merge',
+        owner => $args{owner} // $self->owner,
+        repo  => $args{repo} // $self->repo,
+        id    => $args{id} // $self->number,
+    );
+    $log->infof('URI for PR merge is %s', "$uri");
+    $gh->http_put(
+        uri => $uri,
+        data => {
+            sha => $self->{head}{sha},
+            map { $_ => $args{$_} } grep { exists $args{$_} } qw(
+                commit_title
+                commit_message
+                sha
+                merge_method
+                admin_override
+            )
+        }
+    )
+}
+
+sub Net::Async::Github::PullRequest::cleanup {
+    my ($self, %args) = @_;
+    my $gh = $self->github;
+	$gh->validate_args(%args);
+    die 'invalid owner' if ref $self->owner;
+    die 'invalid repo' if ref $self->repo;
+    die 'invalid id' if ref $self->id;
+    my $uri = $gh->endpoint(
+        'git_refs',
+        category => 'heads',
+        owner    => $args{owner} // $self->{head}{repo}{owner}{login},
+        repo     => $args{repo} // $self->{head}{repo}{name},
+        ref      => $args{ref} // $self->branch_name,
+    );
+    $log->infof('URI for PR delete is %s', "$uri");
+    $gh->http_delete(
+        uri => $uri,
+    )
+}
+
 sub repos {
     my ($self, %args) = @_;
 	if(my $user = delete $args{owner}) {
@@ -223,6 +312,9 @@ sub repos {
 	} else {
         $self->api_get_list(
             endpoint => 'current_user_repositories',
+            endpoint_args => {
+                visibility => $args{visibility} // 'all',
+            },
             class => 'Net::Async::Github::Repository',
         )
     }
@@ -609,6 +701,188 @@ sub http_get {
     })
 }
 
+sub http_delete {
+    my ($self, %args) = @_;
+    my %auth = $self->auth_info;
+
+    if(my $hdr = delete $auth{headers}) {
+        $args{headers}{$_} //= $hdr->{$_} for keys %$hdr
+    }
+    $args{$_} //= $auth{$_} for keys %auth;
+
+    my $uri = delete $args{uri};
+    $log->tracef("DELETE %s { %s }", $uri->as_string, \%args);
+
+    # we never cache deletes
+    $self->http->do_request(
+        method => 'DELETE',
+        uri => $uri,
+        %args,
+    )->then(sub {
+        my ($resp) = @_;
+        $log->tracef("Github response: %s", $resp->as_string("\n"));
+        # If we had ratelimiting headers, apply them
+        for my $k (qw(Reset Limit Remaining)) {
+            if(defined(my $v = $resp->header('X-RateLimit-' . $k))) {
+                my $method = lc $k;
+                $self->core_rate_limit->$method->set_numeric($v);
+            }
+        }
+
+        return Future->done(
+            { },
+            $resp
+        ) if $resp->code == 204;
+        return Future->done(
+            { },
+            $resp
+        ) if 3 == ($resp->code / 100);
+        try {
+            return Future->done(
+                $json->decode(
+                    $resp->decoded_content
+                ),
+                $resp
+            );
+        } catch {
+            $log->errorf("JSON decoding error %s from HTTP response %s", $@, $resp->as_string("\n"));
+            return Future->fail($@ => json => $resp);
+        }
+    })->else(sub {
+        my ($err, $src, $resp, $req) = @_;
+        $log->warnf("Github failed with error %s on source %s", $err, $src);
+        $src //= '';
+        if($src eq 'http') {
+            $log->errorf("HTTP error %s, request was %s with response %s", $err, $req->as_string("\n"), $resp->as_string("\n"));
+        } else {
+            $log->errorf("Other failure (%s): %s", $src // 'unknown', $err);
+        }
+        Future->fail(@_);
+    })
+}
+
+sub http_put {
+    my ($self, %args) = @_;
+    my %auth = $self->auth_info;
+
+    if(my $hdr = delete $auth{headers}) {
+        $args{headers}{$_} //= $hdr->{$_} for keys %$hdr
+    }
+    $args{$_} //= $auth{$_} for keys %auth;
+
+    my $uri = delete $args{uri};
+    my $data = delete $args{data};
+    $log->tracef("PUT %s { %s } <= %s", $uri->as_string, \%args, $data);
+    $data = $json->encode($data) if ref $data;
+    $self->http->PUT(
+        $uri,
+        $data,
+        content_type => 'application/json',
+        %args,
+    )->then(sub {
+        my ($resp) = @_;
+        $log->tracef("Github response: %s", $resp->as_string("\n"));
+        # If we had ratelimiting headers, apply them
+        for my $k (qw(Limit Remaining Reset)) {
+            if(defined(my $v = $resp->header('X-RateLimit-' . $k))) {
+                my $method = lc $k;
+                $self->core_rate_limit->$method->set_numeric($v);
+            }
+        }
+
+        return Future->done(
+            { },
+            $resp
+        ) if $resp->code == 204;
+        return Future->done(
+            { },
+            $resp
+        ) if 3 == ($resp->code / 100);
+        try {
+            return Future->done(
+                $json->decode(
+                    $resp->decoded_content
+                ),
+                $resp
+            );
+        } catch {
+            $log->errorf("JSON decoding error %s from HTTP response %s", $@, $resp->as_string("\n"));
+            return Future->fail($@ => json => $resp);
+        }
+    })->else(sub {
+        my ($err, $src, $resp, $req) = @_;
+        $log->warnf("Github failed with error %s on source %s", $err, $src);
+        $src //= '';
+        if($src eq 'http') {
+            $log->errorf("HTTP error %s, request was %s with response %s", $err, $req->as_string("\n"), $resp->as_string("\n"));
+        } else {
+            $log->errorf("Other failure (%s): %s", $src // 'unknown', $err);
+        }
+        Future->fail(@_);
+    })
+}
+
+sub http_post {
+    my ($self, %args) = @_;
+    my %auth = $self->auth_info;
+
+    if(my $hdr = delete $auth{headers}) {
+        $args{headers}{$_} //= $hdr->{$_} for keys %$hdr
+    }
+    $args{$_} //= $auth{$_} for keys %auth;
+
+    my $uri = delete $args{uri};
+    my $data = delete $args{data};
+    $log->tracef("POST %s { %s } <= %s", $uri->as_string, \%args, $data);
+    $data = $json->encode($data) if ref $data;
+    $self->http->POST(
+        $uri,
+        $data,
+        content_type => 'application/json',
+        %args,
+    )->then(sub {
+        my ($resp) = @_;
+        $log->tracef("Github response: %s", $resp->as_string("\n"));
+        # If we had ratelimiting headers, apply them
+        for my $k (qw(Limit Remaining Reset)) {
+            if(defined(my $v = $resp->header('X-RateLimit-' . $k))) {
+                my $method = lc $k;
+                $self->core_rate_limit->$method->set_numeric($v);
+            }
+        }
+
+        return Future->done(
+            { },
+            $resp
+        ) if $resp->code == 204;
+        return Future->done(
+            { },
+            $resp
+        ) if 3 == ($resp->code / 100);
+        try {
+            return Future->done(
+                $json->decode(
+                    $resp->decoded_content
+                ),
+                $resp
+            );
+        } catch {
+            $log->errorf("JSON decoding error %s from HTTP response %s", $@, $resp->as_string("\n"));
+            return Future->fail($@ => json => $resp);
+        }
+    })->else(sub {
+        my ($err, $src, $resp, $req) = @_;
+        $log->warnf("Github failed with error %s on source %s", $err, $src);
+        $src //= '';
+        if($src eq 'http') {
+            $log->errorf("HTTP error %s, request was %s with response %s", $err, $req->as_string("\n"), $resp->as_string("\n"));
+        } else {
+            $log->errorf("Other failure (%s): %s", $src // 'unknown', $err);
+        }
+        Future->fail(@_);
+    })
+}
+
 sub api_get_list {
     use Variable::Disposition qw(retain_future);
     use Scalar::Util qw(refaddr);
@@ -819,13 +1093,20 @@ sub _add_to_loop {
         $self->{ryu} = Ryu::Async->new
     );
 
-    # Dynamic updates
-    $self->add_child(
-        $self->{ws} = Net::Async::WebSocket::Client->new(
-            on_raw_frame => $self->curry::weak::on_raw_frame,
-            on_frame     => sub { },
-        )
-    );
+}
+
+sub ws {
+    my ($self) = @_;
+    $self->{ws} // do {
+        require Net::Async::WebSocket::Client;
+        $self->add_child(
+            my $ws = Net::Async::WebSocket::Client->new(
+                on_frame => $self->curry::weak::on_frame,
+            )
+        );
+        Scalar::Util::weaken($self->{ws} = $ws);
+        $ws
+    };
 }
 
 1;
@@ -836,5 +1117,5 @@ Tom Molesworth <TEAM@cpan.org>
 
 =head1 LICENSE
 
-Copyright Tom Molesworth 2014-2017. Licensed under the same terms as Perl itself.
+Copyright Tom Molesworth 2014-2018. Licensed under the same terms as Perl itself.
 
